;; $Id: rcs.el,v 1.52 2002/01/27 19:59:44 billl Exp $
;; Copyright (C) 1986, 1987 Edward V. Simpson

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;; vc has the nasty habit of starting up all files write protected
(remove-hook 'find-file-hooks 'vc-find-file-hook) 

(defvar rcs-delta-posthook nil
  "Hook called after completing a delta for co/ci, gets file and version as args")
(defvar rcs-delta-prehook nil
  "Hook called before performing a delta, typically used to set rcs-index-buffer")
(defvar rcs-index-buffer nil
  "Temporary storage for name of the index buffer to use.")
;;; Vars the user doesn't need to know about.

(defvar rcs-mode-map nil)

(defun rcs-mode ()
  "Major mode for doing an rcs check-in.  
This would need to be a minor mode to be used
Calls the value of text-mode-hook then rcs-mode-hook.
Like Text Mode but with these additional comands:

  C-x ?		show this message
  C-c g         rcs-get-file
  C-x C-d       rcs-delta-file
  C-x d         rcs-diffs
"
  (interactive)
  (setq mode-name "RCS")
  (run-hooks 'rcs-mode-hook))

(if rcs-mode-map
    nil
  (setq rcs-mode-map (make-sparse-keymap))
  (define-key rcs-mode-map "\C-x?" 'describe-mode)
  (define-key rcs-mode-map "\C-cg" 'rcs-get-file)     
  (define-key rcs-mode-map "\C-x\C-d" 'rcs-delta-file)
  (define-key rcs-mode-map "\C-xd" 'rcs-diffs))

(defun rcs-lock-file ()
  "Perform rcs lock."
  (interactive)
  (message "Locking file.")
  (call-process "rcs" nil (get-buffer-create "*rcs-out*") nil "-l" (buffer-file-name))
  (pop-to-buffer "*rcs-out*")
  )

(defun rcs-unlock-file ()
  "Perform rcs lock."
  (interactive)
  (message "Unlocking file.")
  (call-process "rcs" nil (get-buffer-create "*rcs-out*") nil "-u" (buffer-file-name))
  (pop-to-buffer "*rcs-out*")
  )

(defun rcs-first-ci ()
  "Perform rcs ci for new file only."
  (interactive)
  (let ((file (file-name-nondirectory (buffer-file-name)))
	(fullfile (buffer-file-name))
	(buf (current-buffer))
	(rbuf (get-buffer-create "*rcs-out*"))
	header)
    (message "Ci %s." file)
    (let (tmp)
      (if (or (file-exists-p (setq tmp (concat "RCS/" file ",v")))
	      (file-exists-p (setq tmp (concat file ",v"))))
	  (error "File exists: %s" tmp)))
    (if (not (file-exists-p "./RCS")) 
	(if (y-or-n-p "Create RCS directory?")
	    (call-process "mkdir" nil nil nil default-directory "RCS")
	  (error "RCS not made, ci not done.")))
    (set-buffer buf)
    (goto-char (point-min)) 
    (setq header (buffer-substring (point) (progn (end-of-line) (point))))
    (if (or (string-match "^[%#]!" header)
	    (string-match "$Id" header)
	    (string-match "$He" header)) ();; have proper RCS header
      (beginning-of-line)
      (if (y-or-n-p "Insert $Id header for RCS? ")
	  (progn
	   (insert (read-string "To be inserted: " (concat "/* $Id" "$ */")) "\n")
	   (write-file fullfile))))
    (call-process "ci" nil rbuf nil "-u" "-t-\"\"" fullfile)
    (call-process "chmod" nil rbuf nil  "+w" fullfile)
    (find-alternate-file fullfile)
    (pop-to-buffer "*rcs-out*")
    ))

(defun rcs-diffs (str)
  "\
Do rcs diff on current file. Default will compare version in buffer 
to most recent version.  1 arg will compare buffer contents to that 
version (NB use '4' not '1.4' for version number).  2 args will compare 
those 2 versions."
  (interactive "sVersions: ")
  (save-buffer)
  (let* (vtmp file v1 v2)
    (cond 
     ;; if a tmp version pick out the version number from the buffer name
     ((and buffer-file-name (string-match "\\.tmp" buffer-file-name))
      (setq file (substring buffer-file-name 0 (match-beginning 0)))
      (if (string-match "[0-9]+$" (buffer-name))
	  (setq vtmp (match-string 0 (buffer-name)) aa vtmp)
	(error "Unable to find version number: give 2 numbers")))
     ;; otherwise this is the file to use
     (buffer-file-name (setq file buffer-file-name))
     ;; unless we're looking at a log buffer
     ((string-match "\\(.+\\)\\(-log\\)\\|\\(-prs\\)" (buffer-name))
      (setq file (substring (buffer-name) 0 (match-end 1))))
     (t (error "Can't find file name!")))
    (and (file-symlink-p file)
	 (message "Linking to %s." (file-symlink-p file))
	 (setq file (file-truename file)))
    ;; parse the string giving the version numbers: 2 numbers
    (if (string-match
	 "\\([0-9]+\\.\\)*\\([0-9]+\\)[ 	]+\\([0-9]+\\.\\)*\\([0-9]+\\)" str)
	(setq v1 (if (match-beginning 1)
		     (substring str (match-beginning 1)(match-end 2))
		   (concat "1." (substring str (match-beginning 2)(match-end 2))))
	      v2 (if (match-beginning 3)
		     (substring str (match-beginning 3)(match-end 4))
		   (concat "1." (substring str (match-beginning 4)(match-end 4))))))
    ;; or 1 number
    (if (string-match "\\([0-9]+\\.\\)*\\([0-9]+\\)[ 	]*" str)
	(setq v1 (if (match-beginning 1)
		     (substring str (match-beginning 1)(match-end 2))
		   (concat "1." (substring str (match-beginning 2)(match-end 2))))))
    ;; if one version was already set use it
    (and (not v2) vtmp (setq v2 (concat "1." vtmp)))
    (pop-to-buffer "rcs-output")(erase-buffer)
    ;; this will generate errors due to the blanks if v1 or v2 don't exist
    (cond 
     ((and v1 v2) (call-process "rcsdiff" nil t nil "-w"
				(concat "-r" v1)(concat "-r" v2) file))
     (v1 (call-process "rcsdiff" nil t nil "-w" (concat "-r" v1) file))
     (v2 (call-process "rcsdiff" nil t nil "-w" (concat "-r" v2) file))
     (t  (call-process "rcsdiff" nil t nil "-w" file)))
    (goto-char (point-min))
    ;; clean up those errors
    (while (re-search-forward
	    "rcsdiff error: RCS/,v: No such file or directory\n" 80 t)
      (delete-region (match-beginning 0) (match-end 0)))))

(defun rcs-get-file (filename vers) 
  "\
Perform rcs get on file giving editable version with name that 
reflects version number.  Enter 'log' as version number to 
get comment listing."
  (interactive "FFile name: \nsVersion number: ")
  (let* ((from-file
	  (file-truename 
	   (expand-file-name
	    (cond
	     ((string-match "\\.tmp$" filename)
	      (substring filename 0 (match-beginning 0)))
	     ((string-match "\\(.+\\)-\\(log\\)\\|\\(prs\\)" (buffer-name))
	      (substring (buffer-name)
			 (match-beginning 1)(match-end 1)))
	     (filename)
	     (t (error "File %s not found." filename))))))
	 (version 
	  (cond 
	   ((not vers) "-r1.")
	   ((string-match "[0-9]+\\." vers)(concat "-r" vers))
	   ((string-match "-[0-9]+" vers)
	    (save-excursion
	      (goto-char (point-min))
	      (or (re-search-forward "$[IH].+,v [0-9]+\\.\\([0-9]+\\)" nil t)
		  (error "Id or Header not found in this file."))
	      (concat 
	       "-r1."
	       (setq vers
		     (int-to-string 
		      (+ (string-to-number vers)
			 (string-to-number 
			  (buffer-substring (match-beginning 1)(match-end 1)))))))))
	   ((concat "-r1." vers))))
	 (ldir (file-name-directory from-file))
	 (filename (reduce-file-name from-file))
	 (rcsid (if (string-match ",v$" filename) 
		    filename (concat "RCS/" filename ",v")))
	 (cbuf (current-buffer))
	 (tempvisit nil) mode)
    (message "co %s version %s ..." filename vers)
    (cond ((or (string= vers "prs") (string= vers "log"))
	   (pop-to-buffer (concat filename "-log")) (erase-buffer)
	   (call-process "rlog" nil t nil rcsid)
	   (goto-char (point-min))
	   (if (re-search-forward "^-" nil t) (delete-region (point-min) (point))
	     (message "Not under RCS.")
	     (erase-buffer)(insert "Not under RCS.")))
	  ((or (string= vers "h"))
	   (pop-to-buffer (concat filename "-head")) (erase-buffer)
	   (call-process "rlog" nil t nil "-h" rcsid)
	   (setq buffer-file-name from-file)
	   (setq tempvisit t))
	  (t
	   (pop-to-buffer 
	    (generate-new-buffer-name
	     (concat filename  (if (string= vers "") "_" vers))))
	   (erase-buffer)
	   (call-process "co" nil t nil "-q" "-p" version rcsid)
	   (setq mode
		 (cdr (assoc (if 
				 (string-match "\\.[a-z0-9_]+\\'" filename)
				 (concat "\\"
					 (substring 
					  filename (match-beginning 0)(match-end 0)) "\\'")
			       ())
			     auto-mode-alist)))
	   (if mode (funcall mode))
	   (if (file-exists-p from-file)
	       (setq buffer-file-name (concat from-file ".tmp"))
	     (write-file from-file))))
    (set-buffer-modified-p nil)
    (setq default-directory ldir)
    (goto-char (point-min))
    (if tempvisit (pop-to-buffer cbuf))))

(defun rcs-delta-file (comment &optional arg)
  "\
Performs both co and ci on current file. Saving if necessary. Writes to index file
Has 2 hooks: rcs-delta-hook called at beginning can be used to set index file.
rcs-delta-posthook called at end gets file and ver number to do postprocessing."
  (interactive "sDELTA COMMENT: ")
  (let* ((file (or buffer-file-name 
		  (substring (buffer-name) 0 (string-match "[0-9]+$" (buffer-name)))))
	(buf (current-buffer))
	(indf (and current-prefix-arg
		   (read-file-name "Index file: " default-directory "" t "")))
	(flags (file-modes file))
	(lnum (count-lines (point-min) (point)))
	(cflag (> (length comment) 0))
	vers vers1 header pt)
    (and (file-symlink-p file)
	 (error "ERROR!!!!: ``%s'' IS A SYMBOLIC LINK." file))
    (if cflag (put-string-on-kill-ring comment))
    (if (and current-prefix-arg (not (string= indf "")))
	(setq rcs-index-buffer 
	      (get-buffer (or (find-buffer-visiting indf)
			       (find-file-noselect indf))))
      (setq rcs-index-buffer nil))  ;; use indf "" as flag for no index file
    (run-hooks 'rcs-delta-prehook)
    (goto-char (point-min)) 
    (setq header (buffer-substring (point) (progn (end-of-line) (point))))
    (if (or 
	 (string-match "^[#%]!" header)
	 (string-match "$Id" header)
	 (string-match "$He" header)) ();; have proper RCS header
      (beginning-of-line)
      (if (y-or-n-p "Insert $Id header for RCS? ")
	  (insert (read-string "To be inserted: " (concat "/* $Id" "$ */")) "\n")))
    (if (buffer-modified-p buf) (write-file file))
    (set-buffer (get-buffer-create "rcs-output")) (erase-buffer)
    (call-process "rcs" nil t nil "-l" file)
    (goto-char (point-min))
    (if (search-forward "error" (point-max) t)
	(error (buffer-substring (point) (progn (end-of-line) (point))))
      (goto-char (setq pt (point-max))))
    (insert (concat comment "\n"))
    (call-process "ci" nil t nil 
		  "-u" (concat "-m" (if cflag comment "\"\"")) file)
    (goto-char pt) (next-line 1)
    (if (not (re-search-forward "[0-9]\\.\\([0-9]+\\)" nil t))
	(progn (pop-to-buffer "rcs-output")
	       (error "ERROR: see 'rcs-output' buffer")))
    (setq vers (buffer-substring (match-beginning 0) (match-end 0)))
    (setq vers1(buffer-substring (match-beginning 1) (match-end 1)))
    (message "version %s" vers)
    (or (string= indf "") (not cflag)
	(index-insert-comment file (concat "(" vers ")") comment))
    (set-buffer buf)
    (set-file-modes file flags)
    (revert-buffer t t)
    (goto-line lnum)
    (or (get-buffer (concat (file-name-nondirectory file) vers1))
	(rename-buffer (concat (file-name-nondirectory file) vers1)))
    (run-hook-with-args 'rcs-delta-posthook (file-name-nondirectory file) vers1)
    ))

(defun rcs-fix-file ()
  "Comparable to 'sccs fix' replace most recent version with modified
  version without changing comment - version number remains the same."
  (interactive)
  (save-window-excursion 
    (let* ((file (or buffer-file-name 
	     (substring (buffer-name) 0 (string-match "1\\." (buffer-name)))))
	   (buf (current-buffer))
	   (flags (file-modes file)) 
	   vers comment vers1)
      (and (buffer-modified-p) (error "Save buffer first."))
      (set-file-modes file 292)  ;; write only
      (call-process "rcsdiff" nil 
        (progn (pop-to-buffer "rcs-output")(erase-buffer)(current-buffer))
		    nil "-w" file)
      (goto-char (point-min)) (delete-other-windows)
      (if (not (yes-or-no-p "Make changes? ")) (set-file-modes file 448)
	(erase-buffer)  
	(call-process "rlog" nil t nil file)
	(beginning-of-buffer)
	(re-search-forward "[0-9]\\.\\([0-9]+\\)" nil)
	(setq vers (buffer-substring (match-beginning 0) (match-end 0)))
	(setq vers1(buffer-substring (match-beginning 1) (match-end 1)))
	(re-search-forward (concat "^revision[ \t]+" vers))
	(forward-line 2) (beginning-of-line)
	(setq comment (buffer-substring (point) (progn (re-search-forward "^---")
						       (forward-line -1)
						       (end-of-line) (point))))
	(message "Vers. %s: %s" vers comment)
	(erase-buffer)
	(call-process "mv" nil nil nil "-f" file (concat file ".tmp"))
	(call-process "rcs" nil t nil "-u" (concat "-o" vers) file)
	(call-process "rcs" nil t nil "-l" file)
	(pop-to-buffer buf) (write-file file)
	(call-process "ci" nil nil nil 
		      "-u" (concat "-m" comment) file)
	(set-file-modes file flags)
	(revert-buffer t t)
	(rename-buffer (concat (file-name-nondirectory file) vers1))
))))

(defun index-insert-comment (file vers comment)
  (let (pt 
	(cline (concat (date-id-string) " " (file-name-nondirectory file)
		       " " vers " " comment)))
    (if (not (rcs-find-index-buffer (expand-file-name file)))
	(message "No index file: %s" cline) 
      (message "Using index buffer %s." (buffer-name))
      (goto-char (setq pt (point-max)))
      (insert cline "\n")
      (neaten-region pt (1- (point)) 2)
      (goto-char (point-max))
      (save-buffer))))

(defun rcs-find-index-buffer (file)
  "Find index file somewhere: 1)user given,2)'index' buffer,3)'index' file"
  (cond 
   ;; with prefix arg may have picked an index file
   (rcs-index-buffer 
    (set-buffer rcs-index-buffer)
    (revert-buffer t t)
    (setq rcs-index-buffer nil) t)
   ;; else find a file named index in the underlying directory
   ;; which may either be in a buffer somewhere or not
   ((file-exists-p (concat (file-name-directory file) "index"))
    (set-buffer
     (or (find-buffer-visiting (concat (file-name-directory file) "index"))
	 (find-file-noselect (concat (file-name-directory file) "index"))))
    (revert-buffer t t))
   (t ())))
  
(defun neaten-line (&optional arg)
  "Do fill-region on a line and indent all but first line by ARG (default 2).
^U by itself is taken as 0."
  (interactive "P")
  (and (listp arg) (setq arg 0)) ;; take care of ^U -> (4) -> 0
  (save-excursion 
    (neaten-region (progn (beginning-of-line)(point))
		  (progn (end-of-line)(point)) (or arg 2))
))

(defun neaten-region (beg end indent)
  "Do fill-region and indent all but first line by ARG."
  (interactive "r\np")
  (save-excursion
    (fill-region beg end)
    (goto-char beg)
    (forward-line)
    (indent-region (point) (1- end) indent)))

(provide 'nrnrcs)
