;; $Id: simctrl.el,v 1.113 2002/04/16 20:53:25 billl Exp $

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.
;; 
;; simctrl package by Bill Lytton - 11/1/93
;;

(require 'nrnrcs)

(defvar sim-default-element ""
  "Current active element for showing parameters etc.")
(defvar sim-output-file ""
  "Name of ouput file for printed data, local to index file only")
(defvar sim-default-directory ()
  "Directory containing index file and various input files.
Executable 'special' for NEURON simulation will be sought first in
sim-default-directory/mod then in sim-default-directory.")
(defvar sim-redirect-output ()
  "*Set to nil for normal operation, is buffer local and should always be nil in
interactive buffer.  If output is being redirected in an undesirable
manner, reset this to nil")
(make-variable-buffer-local 'sim-redirect-output)
(defvar simload-list ()
  "Holds list of files to be opened after simload returns")
(defvar sim-files-alist ()
  "Caches the (file . version) pairs being used in the current simulation.
Local to index buffer only.")
(defvar sim-template-files ()
  "List of template files with current versions in style of sim-files-alist
Local to index buffer only.")
(defvar sim-template-loaded-flag () "*Flag to indicate that templates have been loaded")
(defvar sim-file-open nil
  "Simulator command used to open a file after replacing ## with filename")
(defvar sim-index-buffer nil
"Buffer of index file being used currently")
(defvar sim-buffer nil
"Buffer running the simulation.")
(defvar sim-emacs-buffers nil
"List of buffers that are related to this index file
Local to index buffer only.")
(defvar sim-runnum 0
"Sequential numbering of saved simulations.  Used for label in index file.
Should correspond to variable 'runnum' known to simulator.
Local to index buffer only.")
(defvar sim-library-path (getenv "HOC_LIBRARY_PATH")
  "Search path used to set env HOC_LIBRARY_PATH")
(defvar sim-disp-output-command "xgraph"
  "*Command to run on output file to look at results.")
;; Directory where perl programs needed by simctrl reside.
(setq sim-bin-dir (concat (getenv "SIMCTRL") "/" "bin/"))
;; Regular expression for picking out version numbers of form '[5.93.31]'.
(setq sim-vers-rexp "[[/]\\(\\([0-9]+\\.\\)+[0-9]+\\)[]/]")
;; Regular expression for picking out file names
(setq sim-files-rexp "^@@[ \t]*\\(\\([a-z0-9_\\.]+[ \t]+\\)+[a-z0-9_\\.]+\\)[ \t]*$")
;; permit reading in 2 index buffers based on soft links
(setq find-file-existing-other-name nil) 

(defvar sim-localvar-list 
  '('sim-default-directory 
    'sim-index-buffer      
    'sim-buffer            
    'sim-library-path
    'sim-redirect-output
    )
  "List of sim variables that need to be known by every buffer related
to the process and that don't change.  This means the process buffer,
index buffer, hoc, mod and multi file buffers.")

(defvar sim-indexvar-list 
  '('sim-files-alist       
    'sim-template-files    
    'sim-template-loaded-flag
    'sim-runnum
    'sim-output-file
    'sim-emacs-buffers
    )
"These variables must be global to the buffers governing a
particular simulation.  They are kept as buffer local variables in
the index buffer.  Any routine that needs to access them should
either '(set-buffer sim-index-buffer) first or use get-ilv or
set-ilv with the name as an argument.")

(defsubst sim-proc ()
  "Returns the current nrn process. See variable sim-buffer."
  (or (get-buffer-process sim-buffer)
      (error "No current process. See variable sim-buffer")))

;; utility programs to access local variables maintained in another buffer
(defsubst get-ilv (lv)
"'get index local variable'. Return value of index buffer LOCAL VARIABLE."
  (let (val (cbuf (current-buffer)))
    (set-buffer sim-index-buffer)
    (setq val (eval lv))
    (set-buffer cbuf)
    val))

(defsubst set-ilv (lv val)
"'set index local variable'. Set index buffer LOCAL VARIABLE to VALUE."
  (let ((cbuf (current-buffer)))
    (set-buffer sim-index-buffer)
    (set lv val)
    (set-buffer cbuf)
    val))

;; for debugging
(defun sim-get-locvars ()
"Function used to look at local variables for debugging."
  (interactive)
  (with-output-to-temp-buffer "*locvars*"
    (princ "**** sim-localvar-list ****\n")
    (mapcar (lambda (x) 
	      (print (assoc (eval x) (buffer-local-variables)))) 
	    sim-localvar-list)
    (princ "\n\n**** sim-indexvar-list ****\n")
    (mapcar (lambda (x) 
	      (print (assoc (eval x) (buffer-local-variables)))) 
	    sim-indexvar-list)))

(defun sim-all-processes ()
  "Return all sim-buffers that are active. Bug: doesn't check whether
a particular sim buffer is related to the current sim."
  (let ((bfl (buffer-list)) (found ()))
    (while bfl
      (set-buffer (car bfl))
      (if (and
	   (assoc 'sim-emacs-buffers (buffer-local-variables)) ;; index buffers
	   sim-buffer
	   (buffer-name sim-buffer)
	   (get-buffer-process sim-buffer))
	  (setq found (cons (get-buffer-process sim-buffer) found)))
      (setq bfl (cdr bfl)))
    found))

(defun sim-what-sims (&optional arg)
"Look through all emacs buffers to find index files that are
associated with a simulation.  Produce a brief listing of these.
Prefix arg gives longer listing equivalent to 'sim-portrait' for each one."
  (interactive "P")
  (let ((bfl (buffer-list))
	(sep (concat (make-string 60 ?*) "\n"))
	(dir ""))
    (or arg (message "Use ^U for portrait of each sim."))
    (with-output-to-temp-buffer "*Sim-description*"
      (while bfl
	(set-buffer (car bfl))
	(cond 
	 ((assoc 'sim-emacs-buffers (buffer-local-variables)) ;; index buffers
	  (setq dir 
		(if (file-symlink-p buffer-file-name)
		    (concat (file-name-nondirectory 
			     (substring (file-name-directory buffer-file-name) 0 -1)) "/") ""))

	  (princ 
	   (concat
	    (if arg sep "")
	    (frelname (buffer-file-name sim-index-buffer) "~") ;; file-relative-name
	     "\t("
	    (buffer-name sim-index-buffer)
	    ")\n\tSIM: "
	    (if (not sim-buffer) "none"
	      (let ((str (buffer-name sim-buffer)))
		(if (not (comint-check-proc sim-buffer))
		    (concat str " (dead)")
		  (set-buffer sim-buffer)
		  (setq str (concat str " (" (or nrn-remote-machine system-name) ")"))
		  (set-buffer (car bfl))
		  str)))
	    (if sim-buffer (concat "\n\tOUTPUT: " dir sim-output-file) "\n\n") "\n\n"))
	  (if arg (sim-describe nil))))
	(setq bfl (cdr bfl))))))


;; user description similar to sim-get-locvars (except looks to index file)
(defun sim-portrait (&optional arg)
  "Put the current situation in a temporary buffer. Optional arg gives extra commentary."
  (interactive "P")
  (or (assoc 'sim-index-buffer (buffer-local-variables))
      (error "This buffer does not seem to be associated with a simulation."))
  (or arg (message "Use ^U for portrait with explanation."))
  (with-output-to-temp-buffer "*Sim-description*"
    (sim-describe arg)))

(defun sim-describe (arg)
  "Princ the current situation."
  (let ((al (get-ilv 'sim-files-alist))
	(tp (get-ilv 'sim-template-files))
	(pr (lambda (x) (concat (car x) " v" (cdr x)))))
    (princ 
     (concat "DIRECTORY: " sim-default-directory 
	     (if arg "\t(directory housing simulation files)\n\n" "\n\n")))
    (princ 
     (concat "INDEX:\t file '"
	     (frelname (buffer-file-name sim-index-buffer) sim-default-directory)
	     "' in buffer '" (buffer-name sim-index-buffer) "'"
	     (if arg "\t(master index)\n\n" "\n\n")))
    (princ 
     (concat "SIMULATION IN BUFFER: " 
	     (if sim-buffer
	       (prog2
		   (set-buffer sim-buffer)
		   (concat (buffer-name) " (" (or nrn-remote-machine system-name) ")")
		 (set-buffer sim-index-buffer))
	       "none")
	     (if arg "\t(process buffer)\n\n" "\n\n")))
    (princ
     (concat "TEMPLATE FILES:\t"
	     (if arg "\n  Files that are loaded once at beginning of simulation using 'sim-load-templates'.
  Definition line for these files begins with '%%'.
  Version '0' means look for file on hoc-library-path (see-below)\n\n\t" "\n\t")
	     (mapconcat pr tp "\n\t") "\n\n"))
    (princ
     (concat "SIMULATION FILES:\t"
	     (if arg "\n  Files that are changed frequently and reloaded repeatedly
  Names are given by last '@@' in index file.
  Version numbers are given in '[]' for each simulation.
  See 'sim-load-sim' and 'sim-save-sim'.\n\n\t" "\n\t")
	     (mapconcat pr al "\n\t") "\n\n"))
    (princ 
     (concat "OUTPUT FILE:\t" (get-ilv 'sim-output-file)
	     (if arg "\n  File name made up of date and run id (runnum).\n\n" "\n\n")))
    (princ 
     (concat "HOC LIB PATH:" 
	     (if arg "\n  Used by NEURON commands load_template(), load_func(), load_proc()
  and by sim-load-template with a '0' version.\n\t" "\n\t")
	     (list-to-sepstr "\n\t" sim-library-path) "\n\n"))
    (princ 
     (concat "RELATED FILES IN EMACS BUFFERS:" 
	     (if arg "\n  'rcs-delta-file' for these will appear in the 'index'.
  Killing the 'index' will kill all of these buffers.\n\t" "\n\t")
	     (mapconcat 
	      (lambda (x) (if (not (buffer-name x)) ""
			    (set-buffer x)
			    (if (local-variable-p 'nrnvec-attrnum)
				(concat (file-name-nondirectory (buffer-file-name x))
					"  (ATTRNUM " nrnvec-attrnum ")")
			      (concat
			       (frelname (buffer-file-name x) sim-default-directory)
			       " in " (buffer-name x)))))
	      (get-ilv 'sim-emacs-buffers) "\n\t")
	     "\n\n"))
    ))

(defun sim-ff ()
  "Create an ff buffer to move around from."
  (interactive)
  (let ((ibuf sim-index-buffer)(buf (concat (sim-name) "-ff")))
    (with-output-to-temp-buffer buf
      (mapconcat 
       (lambda (x) (if (not (buffer-name x)) ""
		     (set-buffer x)
		     (if (local-variable-p 'nrnvec-attrnum)
			 (princ (concat (frelname (buffer-file-name x) sim-default-directory)
					"  (ATTRNUM " nrnvec-attrnum ")\n"))
		       (princ (concat
			       (frelname (buffer-file-name x) sim-default-directory)
			       " in " (buffer-name x) "\n")))))
       (get-ilv 'sim-emacs-buffers) "\n\t")
      (princ (buffer-file-name ibuf)))
    (pop-to-buffer buf)
    (goto-char (point-min))
    (insert "@" (get-ilv 'sim-default-directory) "\n")
    (goto-char (point-min))
    (ff-mode)
    (setq buffer-read-only nil)
    (not-modified)
    ))

(defun sim-name ()
  "Find the name of the sim by truncating index file name."
  (string-match "\\(.+\\)-index" (buffer-name sim-index-buffer))
  (match-string 1 (buffer-name sim-index-buffer)))

;; pass a localvar list to another buffer (process buffer, index file,
;; hoc and mod files)
(defun sim-copy-local-variables (list frombuf tobuf)
  "Inherit all local variables in LIST from bindings in buffer FROMBUF
  (usually index file buffer) to TOBUF.  If buf is current buffer and
  local variables do not already exist (use kill-all-local-variables),
  they will be assigned according to the global values.
Program leaves you in buffer tobuf.
Use (defun bl ()(interactive)(setq blv (buffer-local-variables))) to debug."
  (let (varval)
    (mapcar (lambda (var)
	      (let ((evar (eval var)))
		(make-local-variable evar)
		(set-buffer frombuf)
		(setq varval (eval evar));; temp storage for value from frombuf
		(set-buffer tobuf)
		(set evar varval)
		))
	    list)))

(defun sim-menu-bar (map)
  "Put up menu bar."
  (define-key map [menu-bar sim] (cons "Sim" (make-sparse-keymap "Sim")))
  (define-key map [menu-bar sim dout] 
	'("Output to sim buffer" . sim-direct-output))
  (define-key map [menu-bar sim fver] 
	'("Show File Versions" . sim-current-files))
  (define-key map [menu-bar sim fstatus] 
	'("Show File Status..." . sim-rcs-status ))
  (define-key map [menu-bar sim sstatus] 
	'("Show Sim Status" . sim-portrait ))
  (define-key map [menu-bar sim save] 
	'("Save Sim..." . sim-save-sim))
  (define-key map [menu-bar sim fload] 
	'("Load File..." . sim-load-file))
  (define-key map [menu-bar sim sload] ;; avoid xopen so doesn't crash if errors
    '("Load Scene" . (lambda()(interactive)(nrn-load-scene))))
  (define-key map [menu-bar sim tload] 
	'("Load Templates" . sim-load-templates))
  (define-key map [menu-bar sim iload] 
	'("Load Last Saved Sim" . (lambda()(interactive)(sim-load-sim t))))
  (define-key map [menu-bar sim lload] 
	'("Load Current Sim" . sim-load-sim))
  (define-key map [menu-bar sim ver] 
	'("Set File Version..." . sim-set-file-ver))
  (define-key map [menu-bar sim index] 
	'("Read From Index" . sim-read-index-file))
  (define-key map [menu-bar sim recent] 
	'("Read Last Version" .sim-read-recent-versions))
  )

(defun sim-rcs-status (flag)
"Directly check the RCS status of the files being managed for the
simulation (see function sim-current-files to list these files).

Option of 'current' which checks the $Id line in the files in the
current directory, 'heads' which gives the highest version number or
'diffs' which runs rcs-diffs (which see) comparing each current file
with the latest checked-in version.

With argument (ie ^u) can set sim-files-alist to correspond to current
contents of the directory (c) or to heads (h)."
  (interactive "cHeads (h), current (c), diffs (d), check out (r)")
  (let ((cbuf (current-buffer)) 
	(files (alist-cars (get-ilv 'sim-files-alist)))
	prog ins vers)
    (cond ((= flag ?h) (setq prog "r"
			     ins "RCS version heads:\n"))
	  ((= flag ?c) (setq prog "f"
			     ins "RCS current versions:\n"))
	  ((and (or (= flag ?d)(= flag ? )) (eq (current-buffer) sim-index-buffer) 
		(re-search-forward 
		 sim-vers-rexp (prog2 (end-of-line)(point)(beginning-of-line)) t))
	   (setq vers (buffer-substring (match-beginning 1) (match-end 1))
		 prog "d"
		 ins (format "RCS diffs from versions %s:\n\n" vers)))
	  ((or (= flag ?d)(= flag ? )) 
	   (setq vers (mapconcat 'cdr sim-files-alist ".")
		 prog "d"
		 ins "RCS diffs from baseline version:\n\n"))
	  ((= flag ?r)
	   (or (and (eq (current-buffer) sim-index-buffer) 
		    (re-search-forward 
		     sim-vers-rexp (prog2 (end-of-line)(point)(beginning-of-line)) t))
	       (error "Check out from indexed line in index buffer"))
	   (or (y-or-n-p "WARNING: will replace all sim files ")
	       (error "Check out replaces all sim files in current directory."))
	   (setq vers (buffer-substring (match-beginning 1) (match-end 1))
		 prog "n"
		 ins (format "co versions %s:\n\n" vers))
	   (let ((fls files) (vs (string-to-listr "." vers))) 
	     (while fls 
	       (sim-set-file-ver (car fls) (car vs))
	       (setq fls (cdr fls) vs (cdr vs)))))
	  (t (setq prog (concat sim-bin-dir "rcsver")
		   ins "RCS version heads:\n")))
    (pop-to-buffer "*sim-out*")
    (erase-buffer)
    (sim-copy-local-variables sim-localvar-list cbuf (current-buffer))
    (setq default-directory (get-ilv 'default-directory))
    (insert "Files known to emacs:\n" (sim-current-files) "\n\n" ins)
    (apply 'call-process (concat sim-bin-dir "rcsstat") nil t nil prog (or vers "") files)
    (cond (current-prefix-arg 
	   (progn
	     (goto-char (point-min))
	     (re-search-forward sim-vers-rexp (point-max) t)
	     (setq vers (string-to-listr "." (buffer-substring 
					     (match-beginning 1) (match-end 1))))
	     (set-buffer sim-index-buffer)
	     (while files
	       (set-alist-item sim-files-alist (car files) (car vers))
	       (setq files (cdr files) vers (cdr vers)))
	     (set-buffer "*sim-out*")
	     (goto-char (point-max))
	     (insert "\nEmacs files listing updated to:\n" (sim-current-files))))
	  ((= flag ?r)
	   (let ((fls files))
	     (while fls
	       (and (find-buffer-visiting (car fls))
		    (set-buffer (find-buffer-visiting (car fls)))
		    (revert-buffer t t))
	       (setq fls (cdr fls))))))
    (goto-char (point-min))
    (pop-to-buffer cbuf)
    ))

(defun sim-template-status (flag)
  "Check RCS status of templates -- see sim-rcs-status"
  (interactive "cHeads (h), current (c), diffs (d)")
  (let ((cbuf (current-buffer)) 
	(files (alist-cars (get-ilv 'sim-template-files)))
	prog ins vers)
    (cond ((= flag ?h) (setq prog "r"
			     ins "RCS version heads:\n"))
	  ((= flag ?c) (setq prog "f"
			     ins "RCS current versions:\n"))
	  ((or (= flag ?d)(= flag ? ))
	   (setq vers (mapconcat 'cdr sim-template-files ".")
		 prog "d"
		 ins "RCS diffs from baseline version:\n\n")))
    (pop-to-buffer "*sim-out*")
    (erase-buffer)
    (sim-copy-local-variables sim-localvar-list cbuf (current-buffer))
    (setq default-directory (get-ilv 'default-directory))
    (insert ins)
    (apply 'call-process (concat sim-bin-dir "rcsstat") nil t nil prog (or vers "") files)
    (goto-char (point-min))
    (pop-to-buffer cbuf)))

(defun sim-current-files ()
"List the contents of variable sim-files-alist in a readable form.
This is the list of files under sim controls with what emacs believes
to be the current versions.  

See sim-rcs-status to determine the actual contents of the directory,
there should be no discrepency."
  (interactive)
  (let ((str "") (cbuf (current-buffer)) files)
    (set-buffer sim-index-buffer)
    (setq files (alist-cars sim-files-alist))
    (while files
      (setq str (concat str (car files) 
			" " (get-alist-item sim-files-alist (car files)) " "))
      (setq files (cdr files)))
    (setq str (substring str 0 (- (length str) 1))) ;; remove trailing blank
    (set-buffer cbuf)
    (if (interactive-p) (message str)
      str)))

(defun sim-load-index-item ()
  "Replaced with sim-load-sim with prefix argument (^U)"
  (interactive)
  (message   "Replaced with sim-load-sim with prefix argument (^U)"))

(defun sim-reset-file-buffer (file vers)
  "Revert a buffer that has gotten out of sync with version in directory."
  (let ((fullfile (concat sim-default-directory file))
	buf pt)
    (if (setq buf (find-buffer-visiting fullfile))
	(progn 
	  (set-buffer buf)
	  (setq pt (point))
	  (revert-buffer t t)
	  (and (get-buffer (concat file vers)) (kill-buffer (concat file vers)))
	  (rename-buffer (concat file vers))
	  (if (<= pt (point-max))
	      (set-window-point (get-buffer-window buf) pt))))))

(defun sim-load-sim (&optional arg) 
  "Use variable sim-files-alist to load simulation using simload.
Will not overwrite an altered file if it has the requested version
number. Calls program simload in sim-bin-dir.

simload goes through the following flow chart: 
1) does the file exist? 
  no -> check it out, 
  yes -> 
2) has the file been altered since the last sim-save-sim?
  no -> do nothing, 
  yes -> 
3) is the version number in the directory the one we are looking for (ie the one in
sim-files-alist)?
  yes -> do nothing (even if changes have been made), 
  no -> 
4) has the file been changed?
  no -> overwrite it with requested version,
  yes -> ask user the following:
5) What should I do about this changed version? Options:
Ci and use new version (ci), leave (le), mv to .tmp (mv), or discard (rm)?
  ci or le will cause the file present in the directory to be used.
  mv or rm will cause the requested version number to be used."
  (interactive "P")
  (let ((sbuf sim-index-buffer))
    (save-excursion
      (cond (arg (sim-read-index-file)
		 (if (file-exists-p ".last_simsave")
		     (delete-file ".last_simsave"))))
      (cd sim-default-directory)
      (set-buffer sim-index-buffer)
      (make-comint "load-sim" (concat sim-bin-dir "simload") nil
		   (list-to-string (alist-cars sim-files-alist))
		   (list-to-string (alist-cdrs sim-files-alist)))
      (setq simload-list (alist-cars sim-files-alist))
      (set-buffer "*load-sim*")
      (sim-copy-local-variables sim-localvar-list sbuf (current-buffer))
      (setq default-directory (get-ilv 'default-directory))
      (erase-buffer)
      (set-process-filter (get-buffer-process (current-buffer)) 'simload-filter))))

(defun simload-filter (proc string)
  "Filter called to process output from perl script 'simload'."
  ;; modified from emacs-lisp manual p. 618 (33.8.2: ordinary-insertion-filter)
  (let ((inp ""))
    (cond ((string-match "simload done" string)
	   (set-buffer "*load-sim*")
	   (insert string)
	   (sim-update-buffers)
	   (set-process-filter proc nil)
	   (mapcar 'sim-open-file simload-list)
	   (pop-to-buffer "*load-sim*")
	   (pop-to-buffer sim-buffer))
	  (t (pop-to-buffer "*load-sim*")
	     (goto-char (process-mark proc))
	     (insert string)
	     (set-marker (process-mark proc) (point))))
    ))

(defun sim-save-sim (comm)
"RCS any of the files needed for the simulation as defined by
sim-files-alist and put comment in the index file.  Also query the
user as to whether they want to save sim-output-file as an archival file."
  (interactive "sSIM comment: ")
  (let ((fls (alist-cars (get-ilv 'sim-files-alist))) vers changed pt buff
	(sbuf sim-index-buffer)
	(simsave-end-regexp
	 "Replaced: \\(.*\\); Current versions: *\\([0-9\\.]+\\)")
	(line ""))
    (mapcar (lambda (x) 
	      (let ((buf (find-buffer-visiting (concat sim-default-directory x))))
		(if (and buf (buffer-modified-p buf))
		    (progn (put-string-on-kill-ring comm)
			   (error "Unsaved buffer: %s" x))))) fls)
    (set-buffer (get-buffer-create "*save-sim*"))(erase-buffer)
    (sim-copy-local-variables sim-localvar-list sbuf (current-buffer))
    (setq default-directory (get-ilv 'default-directory))
    (call-process (concat sim-bin-dir "simsave") nil t nil (list-to-string fls) comm)
    (goto-char (point-min))
    (insert "Comment: " comm "\n")
    (if (search-forward "Error" (point-max) t)
	(progn (pop-to-buffer "*save-sim*")
	       (message "Save-sim error: see buffer.")))
    (goto-char (point-min))
    (or (re-search-forward simsave-end-regexp (point-max) t)
	(error "Can't find version numbers in *save-sim* buffer."))
    (if (string= ""
		 (setq changed (buffer-substring (match-beginning 1) (match-end 1))))
	(setq changed nil)
      (setq changed (string-to-listr " " changed)))
    (setq vers (buffer-substring (match-beginning 2) (match-end 2)))
    (if changed ()
      (message "WARNING: no files changed (see *save-sim* buffer)")
      (insert 
       "\n" (make-string 60 ?*)
       "\n\nWARNING: possible error."
       "\nIf diffs exist in sim files, need to remove file: '.last_simsave'.\n\n"
       (make-string 60 ?*) "\n")
      (pop-to-buffer "*save-sim*"))
    (while changed
      (and (setq buff (find-buffer-visiting
		       (concat sim-default-directory (car changed))))
	   (set-buffer buff) 
	   (setq pt (point))
	   (revert-buffer t t)
	   (goto-char pt))
      (setq changed (cdr changed)))
    ;; important side effect: move to index buffer to set and output vars
    (setq line (sim-insert-comment vers comm)  
	  sim-runnum (1+ sim-runnum)
	  sim-output-file (format "data/d%s.%02d" (date-id-string) sim-runnum))
    (and (buffer-name sim-buffer) sim-buffer
	 (comint-send-string 
	  (sim-proc) 
	  (format "comment = \"%s\"\nrunnum = %d\noutput_file = \"%s\"\n"
		  line sim-runnum sim-output-file)))
    (if (or (not (file-exists-p sim-output-file))
	    (y-or-n-p "Save data file? ")) ()
      (delete-file sim-output-file)
      (set-buffer "*save-sim*")
      (insert (format "\nremoved %s\n" sim-output-file)))
    (setq vers (string-to-listr "." vers))
    (while fls
      (sim-set-file-ver (car fls) (car vers))
      (setq fls (cdr fls) vers (cdr vers)))
    ))

(defun sim-insert-comment (vers comment)
"Insert a comment into the index file."
  (sim-find-index-file)
  (let ((pt (point-max))
	(line (format "%s.%02d [%s] %s" (date-id-string) sim-runnum vers comment)))
    (goto-char pt)
    (and (file-symlink-p buffer-file-name) ;; grab the directory name
	 (setq line (concat
		     (file-name-nondirectory 
		      (substring (file-name-directory buffer-file-name) 0 -1)) "/" line)))
    (insert line "\n")
    (neaten-region (or pt (point-min)) (1- (point)) 2)
    (goto-char (point-max))
    (write-region (point-min)(point-max) buffer-file-name nil 0)
    (sim-revert-buffers buffer-file-name)
    (message line)
    line))

(defun sim-save-templates (comm)
"RCS template files"
  (interactive "sComment: ")
  (let ((fls (alist-cars (get-ilv 'sim-template-files))) vers changed pt buff
	(sbuf sim-index-buffer)
	(simsave-end-regexp
	 "Replaced: \\(.*\\); Current versions: *\\([0-9\\.]+\\)")
	(line ""))
    (mapcar (lambda (x) 
	      (let ((buf (find-buffer-visiting (concat sim-default-directory x))))
		(if (and buf (buffer-modified-p buf))
		    (progn (put-string-on-kill-ring comm)
			   (error "Unsaved buffer: %s" x))))) fls)
    (set-buffer (get-buffer-create "*save-sim*"))(erase-buffer)
    (sim-copy-local-variables sim-localvar-list sbuf (current-buffer))
    (setq default-directory (get-ilv 'default-directory))
    (and (file-exists-p ".last_simsave")(delete-file ".last_simsave"))
    (call-process (concat sim-bin-dir "simsave") nil t nil (list-to-string fls) comm)
    (goto-char (point-min))
    (insert "Comment: " comm "\n")
    (if (search-forward "Error" (point-max) t)
	(progn (pop-to-buffer "*save-sim*")
	       (message "Save-sim error: see buffer.")))
    (goto-char (point-min))
    (or (re-search-forward simsave-end-regexp (point-max) t)
	(error "Can't find version numbers in *save-sim* buffer."))
    (if (string= ""
		 (setq changed (buffer-substring (match-beginning 1) (match-end 1))))
	(setq changed nil)
      (setq changed (string-to-listr " " changed)))
    (setq vers (buffer-substring (match-beginning 2) (match-end 2)))
    (if changed ()
      (message "WARNING: no files changed (see *save-sim* buffer)")
      (pop-to-buffer "*save-sim*"))
    (while changed
      (and (setq buff (find-buffer-visiting
		       (concat sim-default-directory (car changed))))
	   (set-buffer buff) 
	   (setq pt (point))
	   (revert-buffer t t)
	   (goto-char pt))
      (setq changed (cdr changed)))
    ;; important side effect: move to index buffer to set and output vars
    (setq vers (string-to-listr "." vers))
    (sim-find-index-file)
    (goto-char (point-max))
    (insert
     (format "%%%% %s (%s) %s (%s)\n" (list-to-string fls) 
	     (list-to-sepstr "." vers)
	     comm (date-id-string)))
    (write-region (point-min)(point-max) buffer-file-name nil 0)
    (sim-revert-buffers buffer-file-name)
    (while fls 
      (setcdr (assoc (car fls) sim-template-files) (car vers))
      (setq fls (cdr fls) vers (cdr vers)))
    ))

(defun sim-revert-buffers (filename)
  "Reverts all buffers visiting a file, returns list of buffers."
  (interactive "fFilename: ")
  (let ((truename (abbreviate-file-name (file-truename filename)))
	(list (buffer-list)) (found ()))
    (while list
      (save-excursion
	(set-buffer (car list))
	(if (and buffer-file-name
		 (string= buffer-file-truename truename))
	    (progn
	      (setq found (cons (car list) found))
	      (clear-visited-file-modtime)
	      (insert-file-contents truename nil nil nil t)
	      (set-buffer-modified-p nil)))
	(setq list (cdr list))))
    found))

(defun sim-update-buffers ()
  "Reads output of simload in order to update version nums after ci."
  (let (file fullfile vers buf (cbuf (current-buffer)) (chg-flag nil))
    (goto-char (point-min))
    (while (re-search-forward "^Checking in \\(.+\\)$" (point-max) t)
      (setq file (buffer-substring (match-beginning 1) (match-end 1))
	    fullfile (concat sim-default-directory file)
	    chg-flag t)
      (re-search-forward "^new revision: [0-9]+\\.\\([0-9]+\\);" (point-max) t)
      (setq vers (buffer-substring (match-beginning 1) (match-end 1)))
      (sim-set-file-ver file vers)
      (if (setq buf (find-buffer-visiting fullfile))
	  (progn (set-buffer buf) (rename-buffer (concat file vers) t) 
		 (revert-buffer t t) (set-buffer cbuf))))
    (sim-find-index-file)
    (if (verify-visited-file-modtime (current-buffer)) ()
      (revert-buffer t t)
      (goto-char (point-max)))
    (and chg-flag
	 (or (re-search-backward "([0-9]+\\.[0-9]+)" (point-min) t)
	     (error "Most recent change not found in index file."))
	 (replace-match
	  (concat 
	   "[" (list-to-sepstr "." (alist-cdrs (get-ilv 'sim-files-alist))) "]"))
	 (set-mark (point))
	 (exchange-point-and-mark))
    ))

(defun sim-open-file (file)
  "Use value of sim-file-open to open a file"
  (interactive "FFilename: ")
  (string-match "##" sim-file-open)
  (let ((begstr (substring sim-file-open 0 (match-beginning 0)))
	(endstr (substring sim-file-open 
			   (match-end 0) (length sim-file-open))))
    (comint-send-string (sim-proc) (concat begstr file endstr))))

(defun index-insert-string (str)
  (let (pt fl)
    (if (not (bufferp sim-index-buffer))
	;; will write directly to file in current directory
	(progn (set-buffer (get-buffer-create " indtmp")) (erase-buffer))
      (set-buffer sim-index-buffer) 
      (revert-buffer t t)
      (setq pt (point-max))
      (goto-char pt))
    (insert str)
    (neaten-region (or pt (point-min)) (1- (point)) 2)
    (if (buffer-file-name)
	(save-buffer)
      (if (file-exists-p (setq fl (concat (file-name-directory fl) "index") ))
	  (append-to-file (point-min) (point-max) fl)
	(message "No index file.")))))

(defun sim-index-revert ()
  "Revert the index buffer; typically after writing to it from hoc."
  (interactive)
  (if (not (bufferp sim-index-buffer)) ()
    (set-buffer sim-index-buffer)
    (set-buffer sim-index-buffer) 
    (revert-buffer t t)
    (goto-char (point-max))))

(defun sim-set-known-file (file vers)
  "Set a file only if already in alist"
  (let ((cbuf (current-buffer))
	aelt atlt str)
    (set-buffer sim-index-buffer)
    (setq aelt (assoc file sim-files-alist))
    (setq atlt (assoc file sim-template-files))
    (if (or aelt atlt)
	(progn
	  (goto-char (point-max))      
	  (or (re-search-backward "([0-9]+\\.[0-9]+)" (point-min) t)
	      (error "Most recent change not found in index file."))
	  (cond 
	   (aelt
	    (setcdr aelt vers)
	    (replace-match 
	     (concat "[" 
		     (list-to-sepstr "." (alist-cdrs sim-files-alist)) "]")))
	   (atlt
	    (setcdr atlt vers)
	    (replace-match
	     (concat "(" 
		     (list-to-sepstr "." (alist-cdrs sim-template-files)) ")"))
	    (or (re-search-backward "\\(^[0-9][0-9].....\\) +[a-z0-9_\\.]+ " (point-min) t)
		(error "Err2: recent change not found in index file."))
	    (setq str (concat " (";; the datestr
			      (buffer-substring (match-beginning 1)(match-end 1)) ")"))
	    (replace-match 
	     (concat "%% " (list-to-string (alist-cars sim-template-files)) " "))
	    (goto-char (point-max));; put the datestr at end of the comment
	    (forward-line -1)
	    (end-of-line)
	    (insert str)))
	  (goto-char (point-max))
	  (save-buffer)))
    (set-buffer cbuf)
    ))

(defun sim-set-file-ver (file vers)
  "Set version for given simulation file.
To change files being used need to add a new @@ line to index."
  (interactive 
   (list (completing-read "File name: " (get-ilv 'sim-files-alist) nil t nil nil)
	 (read-string "New version: " nil)))
   (let (aelt (cbuf (current-buffer)))
     (set-buffer sim-index-buffer)
     (setq aelt (assoc file sim-files-alist))
     (if aelt
	 (setcdr aelt vers)
       (message "New file being included in simulations.")
       (setq aelt (cons file vers))
       (setq sim-files-alist (cons aelt sim-files-alist))
       (sim-find-index-file)
       (goto-char (point-max))
       (insert "@@ " (list-to-string (alist-cars sim-files-alist)) "\t"
	       (format "%s" (alist-cdrs sim-files-alist)
		       )))
     (set-buffer cbuf)))

(defun sim-get-file-ver (file)
  "Get version for given file"
  (cdr (assoc file (get-ilv 'sim-files-alist))))

(defun sim-read-index-file ()
  "Look in current index file to set sim-files-alist to filenames and
version numbers."
  (interactive)
  (save-excursion
    (let (files ptr vers temp)
      (if (not (eq (current-buffer) sim-index-buffer))
	  (sim-read-recent-versions)
	;; else assume in index file and use current line versions
	(sim-find-template-vers)
	(beginning-of-line)
	(or (re-search-forward sim-vers-rexp ptr t)
	    (error "Must be on line with [n1.n2. ...]"))
	(setq vers (buffer-substring (match-beginning 1) (match-end 1))
	      ptr (point))
	(or (re-search-backward sim-files-rexp (point-min) t)
	    (error "No file list found."))
	(setq files (buffer-substring (match-beginning 1) (match-end 1)))
	(if (equal (setq temp (sim-build-alist files vers (1+ (point)) ptr))
		   sim-files-alist) ()
	  (message "Changing default files.")
	  (setq sim-files-alist temp))
	))))

(defun sim-read-recent-versions ()
  "Find index file to set sim-files-alist to most recent filenames and
version numbers."
  (interactive)
  (save-excursion
    (let ((runnum-rexp (concat (date-id-string) "\\.\\([0-9][0-9]\\) "))
	  files ptr vers temp)
      (sim-find-index-file)
      (goto-char (point-max))
      (if (not (re-search-backward sim-files-rexp (point-min) t));; look for the file list
	  (message "Can't find file list line starting with @@ in index file.")
	(setq files (buffer-substring (match-beginning 1)(match-end 1)))
	(setq ptr (match-end 1));; save line number where found
	;; look for the last version list after the file list
	(goto-char (point-max))
	(or (re-search-backward sim-vers-rexp ptr t) 
	    (error "Can't find version list eg (num.num. ...) in index file."))
	(setq vers (buffer-substring (match-beginning 1)(match-end 1)))
	(goto-char (point-max))
	(if (re-search-backward runnum-rexp (point-min) t)
	    (setq sim-runnum (1+ (string-to-int 
				  (buffer-substring (match-beginning 1)(match-end 1)))))
	  (setq sim-runnum 1))
	(setq sim-output-file (format "data/d%s.%02d" (date-id-string) sim-runnum))
	(if (setq temp (sim-build-alist files vers ptr (point)))
	    (setq sim-files-alist temp))
	))))

(defun sim-find-index-file ()
  "set buffer to sim-index-buffer or generate error."
  (if (not sim-index-buffer)
      (error "sim-index-buffer %s not found." (buffer-name sim-index-buffer))
    (set-buffer sim-index-buffer)
    (revert-buffer t t)))

(defun sim-build-alist (files vers fptr vptr)
"Takes ' '-sep string FILES giving file list and '.'-sep string VERS with ver
list and makes an alist.  FPTR and VPTR are buffer locations to make nice
error message."
  (let ((alist ()))
    (setq files (string-to-listr " " files))  ;; and split it up
    (setq vers (string-to-listr "."  vers))  ;; and split it up
    ;; should have same number of files and vers
    (if (/= (length files) (length vers))       
	(error "Don't have same number of files (line %d) and vers (line %d)."
	       (count-lines 1 fptr) (count-lines 1 vptr))
      (while files  ;; create the alist
	(setq alist 
	      (append alist (list (cons (car files) (car vers))))
	      files (cdr files) 
	      vers (cdr vers)))
      alist
      )))

(defun sim-find-template-vers ()
  "Look in current index file to set sim-template-files to filenames
  and version numbers."
  (interactive)
  (or (eq (current-buffer) sim-index-buffer) (error "Must run from index file."))
  (save-excursion
    (let ((template-rexp
	   "^%%[ \t]*\\(\\([a-z_][a-z0-9_]*.h*oc[ \t]+\\)+\\)(")
	  (tempvers-rexp "(\\(\\([0-9]+\\.\\)*[0-9]+\\))")
	  files ptr vers temp)
      (end-of-line) (setq ptr (point))
      (cond 
       ((not (re-search-backward "^%%\\s +" (point-min) t))
	(setq sim-template-files nil))
       ((re-search-forward "%%\\s +no\\s +templates"  ptr t)
	(setq sim-template-files nil))
       ((not (re-search-forward template-rexp ptr t))
	(setq sim-template-files nil))
       (t
	(setq files (buffer-substring (match-beginning 1) (match-end 1)))
	(end-of-line) (setq ptr (point)) (beginning-of-line)
	(or (re-search-forward tempvers-rexp ptr t)
	    (error "Template file version numbers not found."))
	(setq vers (buffer-substring (match-beginning 1) (match-end 1)))
	(setq sim-template-files (sim-build-alist files vers (point) ptr))
	)
       ))))

(defun sim-load-templates ()
  "Load template files in order given in sim-template-files. See sim-load-sim.
For templates can use 0 as a version number, meaning current version
in main directory.  This will search first CWD then sim-library-path for this file."
  (interactive)
  (if (not (eq (current-buffer) sim-index-buffer)) ()
    (save-excursion
      (or (re-search-backward "^%%" (point-min) t) 
	  (error "Can't find template line beginning with '%%'."))
      (sim-find-template-vers)))
  (if (get-ilv 'sim-template-loaded-flag) 
      (error "Templates already loaded for this buffer")
    (set-ilv 'sim-template-loaded-flag t))
  (let ((tmplist (get-ilv 'sim-template-files))
	(filefound "")
	(sbuf sim-index-buffer)
	fls vers hocpath file fullfile version)
    (if (not tmplist) (message "NO TEMPLATES")
      (setq simload-list ());; global for passing on to filter function
      (while tmplist
	(setq file (car (car tmplist))
	      version (cdr (car tmplist))
	      tmplist (cdr tmplist));; march through tmplist looking for 0
	(if (not (= (string-to-int version) 0));; simload will have to load this one
	    (setq fls (cons file fls);; regular file with version number
		  vers (cons version vers)
		  simload-list (cons file simload-list))
	  (if (file-exists-p file);; else look for the file
	      (setq simload-list (cons file simload-list))
	    (setq hocpath sim-library-path)
	    (while 
		(and hocpath
		     (not (file-exists-p 
			   (setq fullfile
				 (expand-file-name
				  (concat (car hocpath) "/" file))))))
	      (setq hocpath (cdr hocpath)))
	    (if hocpath
		(setq simload-list 
		      (cons fullfile simload-list))
	      (error "File %s not found on HOC_LIBRARY_PATH" file)))
	  (setq filefound;; create message for informing user
		(concat filefound 
			(format "Loading %s directly.\n" fullfile)))))
      (setq simload-list (reverse simload-list))
      (cd sim-default-directory)
      ;; remove last_simsave which is only relevant for sim files anyway
      (and (file-exists-p ".last_simsave")(delete-file ".last_simsave"))
      (make-comint "load-sim" (concat sim-bin-dir "simload") nil
		   (list-to-string (reverse fls))
		   (list-to-string (reverse vers)))
      (set-buffer "*load-sim*")
      (erase-buffer)
      (insert filefound)
      (sim-copy-local-variables sim-localvar-list sbuf (current-buffer))
      (setq default-directory (get-ilv 'default-directory))
      (set-process-filter (get-buffer-process (current-buffer)) 'simload-filter)
      (and (file-exists-p ".last_simsave")(delete-file ".last_simsave")))))

(defun sim-switch-to-buffer (arg)
  "Switch to a file being managed as part of this simulation.  With argument, uses same window"
  (interactive "P")
  (let ((choice-alist
	 (mapcar (lambda (x) (cons (buffer-name x) x))
		 (append (list sim-index-buffer sim-buffer)
			 (get-ilv 'sim-emacs-buffers))))
	(switch (cond ((not arg) 'pop-to-buffer)
		      ((= (car arg) 4) 'switch-to-buffer )
		      ((= (car arg) 16) (lambda(x)(pop-to-buffer x)(other-window 1)))
		      (t (lambda(x)(pop-to-buffer x)(other-window 1)))))
	choice c2)
    (setq choice
	  (completing-read 
	   (format "Simulation buffer (DEFAULT '%s'): "
		   (if (and sim-buffer (not (equal sim-buffer (current-buffer))))
		       (buffer-name sim-buffer)
		     (buffer-name sim-index-buffer)))
	   choice-alist))
    (if (string= choice "")
	(cond
	 ((equal sim-buffer (current-buffer)) (funcall switch sim-index-buffer))
	 ((and sim-buffer (buffer-name sim-buffer)) (funcall switch sim-buffer))
	 ((and sim-index-buffer (buffer-name sim-index-buffer)) (funcall switch sim-index-buffer))
	 (t (error "No current index buffer. Run sim-portrait.")))
      (if (setq c2 (assoc choice choice-alist))
	  (funcall switch (cdr c2))
	(while (and choice-alist 
		    (or
		     (not (car (car choice-alist))) ;; skip nils
		     (not (string-match choice (car (car choice-alist)))))) ;; stop if matchs
	  (setq choice-alist (cdr choice-alist)))
	(if choice-alist
	    (funcall switch (cdr (car choice-alist)))
	  (error "Not a simulation file choice - use completions."))))
    ))

(defun sim-disp-output (comm args)
  "Run shell COMMMAND with ARGS on output file to look at results."
  (interactive (list (read-string "Command: " sim-disp-output-command)
		     (read-string "Args: " (get-ilv 'sim-output-file))))
  (apply 'start-process comm (get-buffer-create "*sim-out*")
	 comm (string-to-listr " " args)))

(defun sim-output-other-window ()
  (interactive)
  (sim-set-sim-buffer (prog2 (other-window 1)(buffer-name)(other-window 1))))

(defun sim-set-sim-buffer (str)
  "Allows you to set the sim-buffer to a given buffer by name to
  facilitate debugging."
  (interactive "bBuffer name: ")
  (let ((bf (get-buffer str)))
    (setq sim-buffer bf)
    (set-ilv 'sim-buffer bf)
    (set-buffer bf)
    (setq sim-buffer bf)))

(defun sim-unset-sim-buffer ()
  "Return variable sim-buffer to nil."
  (interactive)
  (setq sim-buffer nil)
  (set-ilv 'sim-buffer nil))

(defun sim-direct-output (num)
"Use this if output from neuron is being redirected in undesirable ways."
  (interactive "sNum: ")
  (if (string= num "")(setq sim-redirect-output nil)
    (setq sim-redirect-output (string-to-int num))))

(defun sim-grep (str)
  "Grep in all of the sim files"
  (interactive "sString: ")
  (grep (concat "grep -n " str " " 
		(list-to-string (mapcar (lambda (x) (car x)) 
					(get-ilv 'sim-files-alist))))))

(defun sim-move-output ()
  "Move an output file to file by another name"
  (interactive)
  (let* ((rnum (get-ilv 'sim-runnum))
	 (oldf (format "data/v%s.%02d" (date-id-string) (- rnum 2)))
	 (newf (format "data/v%s.%02d" (date-id-string) (- rnum 1))))
    (setq oldf (read-file-name "Saved file: " "" nil t oldf)
	  newf (read-file-name "Rename to: " "" nil nil newf))
    (cond 
     ((string= newf "nil") (delete-file oldf) (message "Removed %s" oldf))
     ((and (file-exists-p newf) (not (y-or-n-p (concat "Overwrite " newf "? ")))) ())
     (t (message "Moving %s to %s" oldf newf)
	(call-process "/usr/bin/mv" nil nil nil oldf newf)))))

(provide 'simctrl)
